@@ -21,8 +21,11 @@
 
 #include <sys/param.h>
 
-#if __FreeBSD_version >= 1200000
-#error This driver does not support FreeBSD 12.x/-CURRENT!
+#if __FreeBSD_version >= 1300000
+#error This driver does not support FreeBSD 13.x/-CURRENT!
+#endif
+#if __FreeBSD_version >= 1200000 && __FreeBSD_version < 1200086
+#error This driver requires FreeBSD 12.0-RC3 or later!
 #endif
 #if __FreeBSD_version >= 1100000 && __FreeBSD_version < 1100122
 #error This driver requires FreeBSD 11.0 or later!
@@ -109,7 +112,9 @@
 #define suser(_td) priv_check((_td), PRIV_DRIVER)
 #endif
 
-#if __FreeBSD_version >= 900041
+#if __FreeBSD_version >= 1100012
+#include <sys/capsicum.h>
+#elif __FreeBSD_version >= 900041
 #include <sys/capability.h>
 #if __FreeBSD_version < 1000053
 #define cap_rights_init(rights, cap) (cap)
@@ -398,6 +403,21 @@ NvS32  nv_free_system_pages  (nv_state_t *, void *);
 #define kmem_arena kernel_map
 #endif
 
+#if __FreeBSD_version >= 1200083
+#define NV_KMEM_ALLOC_CONTIG(size, flags, low, high, align, boundary, memattr) \
+    kmem_alloc_contig(size, flags, low, high, align, boundary, memattr)
+
+#define NV_KMEM_FREE(address, size) \
+    kmem_free(address, size)
+
+#else
+#define NV_KMEM_ALLOC_CONTIG(size, flags, low, high, align, boundary, memattr) \
+    kmem_alloc_contig(kmem_arena, size, flags, low, high, align, boundary, memattr)
+
+#define NV_KMEM_FREE(address, size) \
+    kmem_free(kmem_arena, address, size)
+#endif
+
 #if __FreeBSD_version >= 800000
 #define NV_VM_PAGER_ALLOCATE(type, handle, size, prot, off, cred) \
   vm_pager_allocate(type, handle, size, prot, off, cred)
