@@ -61,6 +61,9 @@ NV_STATUS nvidia_get_rm_ops(nvidia_modeset_rm_ops_t *rm_ops)
 {
     const nvidia_modeset_rm_ops_t local_rm_ops = {
         .version_string = NV_VERSION_STRING,
+        .system_info    = {
+            .allow_write_combining = NV_FALSE,
+        },
         .alloc_stack    = nvidia_modeset_rm_ops_alloc_stack,
         .free_stack     = nvidia_modeset_rm_ops_free_stack,
         .enumerate_gpus = NULL,
