@@ -364,7 +364,7 @@ int nvidia_get_card_info(void *args, int size)
     }
 
     ci = args;
-    memset(ci, 0, sizeof(ci));
+    memset(ci, 0, size);
 
     for (i = 0; i < NV_MAX_DEVICES; i++) {
         sc = devclass_get_softc(nvidia_devclass, i);
@@ -952,27 +952,96 @@ NV_STATUS NV_API_CALL nv_free_kernel_mapping(
     return NV_OK;
 }
 
-NV_STATUS NV_API_CALL nv_create_user_mapping_context(
+static nvidia_filep_t* nv_get_file_private(NvU32 fd, nv_state_t *nv)
+{
+    struct file *pFile = NULL;
+    NV_STATUS status = EINVAL;
+    nvidia_filep_t *nvfp = NULL;
+    struct cdev_privdata *p;
+
+#if __FreeBSD_version >= 1000053
+    cap_rights_t rights;
+#endif
+    struct thread *td;
+
+    td = curthread;
+
+    status = fget(td, fd, cap_rights_init(&rights, CAP_READ), &pFile);
+    if (status != 0)
+        return NULL;
+
+    p = pFile->f_cdevpriv;
+    if(!p)
+        goto done;
+
+    if (NV_IS_CTL_DEVICE(nv))
+    {
+        if (p->cdpd_dtr != nvidia_ctl_dtor)
+            goto done;
+    }
+    else
+    {
+        if (p->cdpd_dtr != nvidia_dev_dtor)
+            goto done;
+    }
+
+    nvfp = (nvidia_filep_t*)p->cdpd_data;
+
+done:
+    /*
+     * If we reach here, pFile is a valid struct file pointer, returned
+     * by fget(9).
+     *
+     * fget(9) incremented the struct file's reference count, which
+     * needs to be balanced with a call to fdrop(9).  It is safe to
+     * decrement the reference count before returning filp->private_data
+     * because we are holding the GPUs lock which prevents freeing the file out.
+     */
+    fdrop(pFile, td);
+
+    return nvfp;
+}
+
+NV_STATUS NV_API_CALL nv_add_mapping_context_to_file(
     nv_state_t *nv,
-    NvU64 address,
-    NvU64 size,
+    nv_usermap_access_params_t* nvuap,
     NvU32 prot,
     void *pAllocPriv,
     NvU64 pageIndex,
-    void *file,
-    void **ppPrivate
+    NvU32 fd
 )
 {
-    *ppPrivate = NULL;
-    return NV_OK;
-}
+    nv_alloc_mapping_context_t *nvamc = NULL;
+    nvidia_filep_t *nvfp = NULL;
 
-void NV_API_CALL nv_destroy_user_mapping_context(
-    nv_state_t *nv,
-    void *pFile,
-    void *pPrivate
-)
-{
+    /* Get the nvidia private file data from file descriptor */
+    nvfp = nv_get_file_private(fd, nv);
+    if (nvfp == NULL)
+        return NV_ERR_INVALID_ARGUMENT;
+
+    nvamc = &nvfp->mmap_context;
+
+    if (nvamc->bValid)
+        return NV_ERR_STATE_IN_USE;
+
+    if (NV_IS_CTL_DEVICE(nv))
+    {
+        nvamc->alloc = pAllocPriv;
+        nvamc->page_index = pageIndex;
+    }
+    else
+    {
+        nvamc->mmap_start = nvuap->mmap_start;
+        nvamc->mmap_size = nvuap->mmap_size;
+        nvamc->access_start = nvuap->access_start;
+        nvamc->access_size = nvuap->access_size;
+        nvamc->remap_prot_extra = nvuap->remap_prot_extra;
+    }
+
+    nvamc->prot = prot;
+    nvamc->bValid = NV_TRUE;
+
+    return NV_OK;
 }
 
 NV_STATUS NV_API_CALL nv_alloc_user_mapping(
@@ -1058,7 +1127,7 @@ NvS32 nv_alloc_contig_pages(
         return ENOMEM;
     }
 
-    address = kmem_alloc_contig(kmem_arena, size, flags, 0,
+    address = NV_KMEM_ALLOC_CONTIG(size, flags, 0,
             sc->dma_mask, PAGE_SIZE, 0, attr);
     if (!address) {
         status = ENOMEM;
@@ -1107,7 +1176,7 @@ failed:
         os_flush_cpu_cache();
 
     if (at->pte_array[0].virtual_address != NULL) {
-        kmem_free(kmem_arena, at->pte_array[0].virtual_address, at->size);
+        NV_KMEM_FREE(at->pte_array[0].virtual_address, at->size);
         malloc_type_freed(M_NVIDIA, at->size);
     }
 
@@ -1133,7 +1202,7 @@ NvS32 nv_free_contig_pages(
     if (at->attr != VM_MEMATTR_WRITE_BACK)
         os_flush_cpu_cache();
 
-    kmem_free(kmem_arena, at->pte_array[0].virtual_address, at->size);
+    NV_KMEM_FREE(at->pte_array[0].virtual_address, at->size);
     malloc_type_freed(M_NVIDIA, at->size);
 
     free(at->pte_array, M_NVIDIA);
@@ -1196,7 +1265,7 @@ NvS32 nv_alloc_system_pages(
     }
 
     for (i = 0; i < count; i++) {
-        address = kmem_alloc_contig(kmem_arena, PAGE_SIZE, flags, 0,
+        address = NV_KMEM_ALLOC_CONTIG(PAGE_SIZE, flags, 0,
                 sc->dma_mask, PAGE_SIZE, 0, attr);
         if (!address) {
             status = ENOMEM;
@@ -1250,7 +1319,7 @@ failed:
     for (i = 0; i < count; i++) {
         if (at->pte_array[i].virtual_address == 0)
             break;
-        kmem_free(kmem_arena, at->pte_array[i].virtual_address, PAGE_SIZE);
+        NV_KMEM_FREE(at->pte_array[i].virtual_address, PAGE_SIZE);
         malloc_type_freed(M_NVIDIA, PAGE_SIZE);
     }
 
@@ -1279,7 +1348,7 @@ NvS32 nv_free_system_pages(
         os_flush_cpu_cache();
 
     for (i = 0; i < count; i++) {
-        kmem_free(kmem_arena, at->pte_array[i].virtual_address, PAGE_SIZE);
+        NV_KMEM_FREE(at->pte_array[i].virtual_address, PAGE_SIZE);
         malloc_type_freed(M_NVIDIA, PAGE_SIZE);
     }
 
@@ -1396,6 +1465,7 @@ int nvidia_mmap_ctl_single(
     struct nvidia_softc *sc = nv->os_state;
     NvU64 pageIndex;
     NvU32 prot;
+    nv_alloc_mapping_context_t *mmap_context;
 
     sp = sc->sp[NV_DEV_STACK_API];
 
@@ -1403,13 +1473,23 @@ int nvidia_mmap_ctl_single(
     if (rmStatus != NV_OK)
         return EAGAIN;
 
-    rmStatus = rm_validate_mmap_request(sp, nv, filep, *offset,
-            size, &prot, (void **)&at, &pageIndex);
-    if (rmStatus != NV_OK) {
+    mmap_context = &filep->mmap_context;
+
+    /*
+     * If no mmap context exists on this file descriptor, this mapping wasn't
+     * previously validated with the RM so it must be rejected.
+     */
+    if (!mmap_context->bValid)
+    {
+        nv_printf(NV_DBG_ERRORS, "NVRM: VM: invalid mmap\n");
         rm_release_api_lock(sp);
-        return EINVAL;
+        return -EINVAL;
     }
 
+    at = (struct nvidia_alloc *)mmap_context->alloc;
+    pageIndex = mmap_context->page_index;
+    prot = mmap_context->prot;
+
     vm_object_reference(at->object);
     *object = at->object;
     *offset = (pageIndex * PAGE_SIZE);
@@ -1427,19 +1507,21 @@ int nvidia_mmap_dev_single(
     vm_object_t *object
 )
 {
-    nvidia_stack_t *sp;
-    NV_STATUS rmStatus;
     struct nvidia_softc *sc = nv->os_state;
-    NvU32 prot;
+    nv_alloc_mapping_context_t *mmap_context = &filep->mmap_context;
 
-    sp = sc->sp[NV_DEV_STACK_API];
+    /*
+     * If no mmap context is valid on this file descriptor then this mapping
+     * wasn't previously validated with the RM so it must be rejected.
+     */
+    if (!mmap_context->bValid)
+    {
+        nv_printf(NV_DBG_ERRORS, "NVRM: VM: invalid mmap\n");
+        return -EINVAL;
+    }
 
-    NV_PCI_CHECK_CONFIG_SPACE(sp, nv, TRUE, TRUE, TRUE);
-
-    rmStatus = rm_validate_mmap_request(sp, nv, filep, *offset, size,
-            &prot, NULL, NULL);
-    if (rmStatus != NV_OK)
-        return EINVAL;
+    *offset = mmap_context->mmap_start;
+    size = mmap_context->mmap_size;
 
     if (IS_UD_OFFSET(nv, *offset, size)) {
         *object = sc->UD_object;
@@ -1518,7 +1600,7 @@ void NV_API_CALL nv_set_dma_address_size(
 #endif
 }
 
-void* NV_API_CALL nv_get_adapter_state(
+nv_state_t* NV_API_CALL nv_get_adapter_state(
     NvU32 domain,
     NvU8  bus,
     NvU8  slot
@@ -1536,18 +1618,18 @@ void* NV_API_CALL nv_get_adapter_state(
 
         if ((nv->pci_info.domain == domain) &&
             (nv->pci_info.bus == bus) && (nv->pci_info.slot == slot)) {
-            return (void *) nv;
+            return nv;
         }
     }
 
-    if (NV_PCI_MATCH_CTL_DEVICE(domain, bus, slot)) {
-        nv = &nvidia_ctl_state;
-        return (void *) nv;
-    }
-
     return NULL;
 }
 
+nv_state_t* NV_API_CALL nv_get_ctl_state(void)
+{
+    return &nvidia_ctl_state;
+}
+
 void NV_API_CALL nv_verify_pci_config(
     nv_state_t *nv,
     BOOL check_the_bars
@@ -1585,6 +1667,7 @@ NV_STATUS NV_API_CALL nv_dma_map_alloc(
     nv_state_t *nv,
     NvU64       page_count,
     NvU64      *va_array,
+    NvBool      contig,
     void      **priv
 )
 {
@@ -1612,7 +1695,15 @@ NV_STATUS NV_API_CALL nv_dma_map_peer(
     return NV_ERR_NOT_SUPPORTED;
 }
 
-NV_STATUS NV_API_CALL nv_dma_unmap_peer(
+void NV_API_CALL nv_dma_unmap_peer(
+    nv_state_t *nv,
+    NvU64       page_count,
+    NvU64       va
+)
+{
+}
+
+NV_STATUS NV_API_CALL nv_dma_map_mmio(
     nv_state_t *nv,
     NvU64       page_count,
     NvU64      *va
@@ -1621,6 +1712,14 @@ NV_STATUS NV_API_CALL nv_dma_unmap_peer(
     return NV_ERR_NOT_SUPPORTED;
 }
 
+void NV_API_CALL nv_dma_unmap_mmio(
+    nv_state_t *nv,
+    NvU64       page_count,
+    NvU64       va
+)
+{
+}
+
 NV_STATUS NV_API_CALL nv_log_error(
     nv_state_t *nv,
     NvU32       error_number,
@@ -1680,8 +1779,8 @@ NV_STATUS NV_API_CALL nv_unregister_user_pages(
 
 NV_STATUS NV_API_CALL nv_get_device_memory_config(
     nv_state_t *nv,
-    NvU32 *pAddr,
-    NvU32 *pAddrGuest,
+    NvU32 *pAddrSysPhys,
+    NvU32 *pAddrGuestPhys,
     NvU32 *pAddrWidth,
     NvU32 *pGranularity,
     NvS32 *pNodeId
@@ -1689,3 +1788,61 @@ NV_STATUS NV_API_CALL nv_get_device_memory_config(
 {
     return NV_ERR_NOT_SUPPORTED;
 }
+
+NV_STATUS NV_API_CALL nv_get_usermap_access_params(
+    nv_state_t *nv,
+    nv_usermap_access_params_t *nvuap
+)
+{
+    return NV_OK;
+}
+
+NV_STATUS NV_API_CALL nv_register_peer_io_mem(
+    nv_state_t *nv,
+    NvU64      *phys_addr,
+    NvU64       page_count,
+    void      **priv_data
+)
+{
+    return NV_OK;
+}
+
+void NV_API_CALL nv_unregister_peer_io_mem(
+    nv_state_t *nv,
+    void       *priv_data
+)
+{
+}
+
+NV_STATUS NV_API_CALL nv_get_ibmnpu_genreg_info(
+    nv_state_t *nv,
+    NvU64      *base,
+    NvU64      *size
+)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+NV_STATUS NV_API_CALL nv_get_ibmnpu_relaxed_ordering_mode(
+    nv_state_t *nv,
+    NvBool *mode
+)
+{
+    return NV_ERR_NOT_SUPPORTED;
+}
+
+void NV_API_CALL nv_register_backlight(
+    nv_state_t *nv,
+    NvU32 displayId,
+    NvU32 currentBrightness
+)
+{
+    /* not implemented */
+}
+
+void NV_API_CALL nv_unregister_backlight(
+    nv_state_t *nv
+)
+{
+    /* not implemented */
+}
